import numpy as np
import numpy.matlib as matlib
import soundfile as sf
import matplotlib.pyplot as plt
from os import walk
import os
import math
import python_speech_features as psf
import webrtcvad

from gmm_support_solution import EM
from gmm_support_solution import classify_probability

def loadDataFromFolder(pcFolderName, bAppendDataToOneVector):
    # Walk through folder and get all filenames
    ppcFilenames = []
    for (_, _, filenames) in walk(os.getcwd() + '\\database\\' + str(pcFolderName) + '\\'):
        ppcFilenames.extend(filenames) 
        break

    if bAppendDataToOneVector:
        # Append all files to one big vector
        pfAllData = np.array([])
        for pcFilename in ppcFilenames:
            pfData, _ = sf.read(os.getcwd() + '\\database\\' + str(pcFolderName) + '\\' + pcFilename)
            pfAllData = np.concatenate((pfAllData, pfData))

        return pfAllData
    else:
        # Return files as matrix
        ppfAllData = []
        for pcFilename in ppcFilenames:
            pfData, _ = sf.read(os.getcwd() + '\\database\\' + str(pcFolderName) + '\\' + pcFilename)
            ppfAllData.append(pfData)

        return np.array(ppfAllData), ppcFilenames

def extractVoiceOnlyData(fSamplerate, fVadFrameLengthInMS, pfAllData):
    # Split complete data in frames of fVadFrameLengthInMS
    iVadSamplesPerFrame = int(fSamplerate * (fVadFrameLengthInMS / 1000))
    ppfAllData = np.array(np.split(pfAllData[:int(pfAllData.shape[0] / iVadSamplesPerFrame) * iVadSamplesPerFrame], int(pfAllData.shape[0] / iVadSamplesPerFrame)))

    # Check if frame contains speech
    pbAllDataIsSpeech = np.zeros(ppfAllData.shape[0], dtype=bool)
    for iFrame in np.arange(ppfAllData.shape[0]):
        pbAllDataIsSpeech[iFrame] = sVad.is_speech(ppfAllData[iFrame,:], 16000)
    
    # Reconstruct speech signal with only speech
    pfOnlySpeech = ppfAllData[pbAllDataIsSpeech,:].flatten()

    return pfOnlySpeech

def splitTrainingData(ppfSig, fVadFrameLengthInMS, fLengthOfTrainingDataInSec):
    # Create array for training data
    iNumberOfTrainingFrames = int(fLengthOfTrainingDataInSec / (fVadFrameLengthInMS / 1000 / 2) )
    ppfTrainingData = np.zeros((iNumberOfTrainingFrames, ppfSig.shape[1]))

    # Fill training data with mfccs from random frames
    piUsedFrames = []
    for iFrame in np.arange(iNumberOfTrainingFrames):
        # Make sure, we use new random frame
        while True:
            iRand = np.random.randint(0,ppfSig.shape[0]-1)
            if iRand not in piUsedFrames:
                piUsedFrames.append(iRand)
                break
        
        # Copy frame into training data set
        ppfTrainingData[iFrame,:] = ppfSig[iRand,:]

    return ppfTrainingData


### Set parameters
fLengthOfTrainingDataInSec = 1       # [0.5, ..., 100]
fLengthOfTestDataInSec     = 1       # [0.5, ..., 100]
iNumberOfMfccs             = 13       # [2, ..., 26]
iSeed                      = 42      # [None, 0, 1, ...]
iNumGaussiansOfGmm         = 1       # [1, ..., 5]
iMaxIterationsOfGmm        = 10      # [1, ...]
bPlotFeatureExample        = True    # [True, False]


### Initialize stuff (fixed values)
iNumSpeakers = 4
fSamplerate = 16000
fVadFrameLengthInMS = 20
pppfTrainingData = []
sVad = webrtcvad.Vad()
pcFolderNames = [str(x + 1) for x in range(iNumSpeakers)]
pcFolderNames.append('ubm')


### Train models
# Set seed for random number generator if seed was set
if iSeed:
    np.random.seed(iSeed)

# Get training data for each speaker and the universal background model
for pcFolderName in pcFolderNames:
    # Load files
    pfAllData = loadDataFromFolder(pcFolderName, True)
    
    # Use a voice activity detection and remove unvoiced frames
    pfOnlySpeech = extractVoiceOnlyData(fSamplerate, fVadFrameLengthInMS, pfAllData)

    # Split speech signal into frames and calculate mfccs of each frame
    ppfMfcc = psf.mfcc(pfOnlySpeech, fSamplerate, numcep=iNumberOfMfccs, winlen=fVadFrameLengthInMS / 1000, winstep=fVadFrameLengthInMS / 1000 / 2)

    # Get random frames with right length
    pppfTrainingData.append(splitTrainingData(ppfMfcc, fVadFrameLengthInMS, fLengthOfTrainingDataInSec))

# Convert list to array
pppfTrainingData = np.array(pppfTrainingData)

### Train GMM with EM-Algortihm
list_models = []
for iModel in np.arange(pppfTrainingData.shape[0]):
    print('Train model {}:'.format(iModel+1))

    # Train model with given training data with em-algortihm
    list_models.append(EM(pppfTrainingData[iModel,:,:], iNumGaussiansOfGmm, iMaxIterationsOfGmm))


### Test unknown data
# Load files
ppfUnknownData, ppcFilenames = loadDataFromFolder('unknown', False)
pppfUnknownData = []
piClassifiedData = []

# Loop through all test datasets and extract probabilities
for iUnknownData in np.arange(len(ppfUnknownData)):
    # Use a voice activity detection and remove unvoiced frames
    pfOnlySpeech = extractVoiceOnlyData(fSamplerate, fVadFrameLengthInMS, ppfUnknownData[iUnknownData])
    
    # Calculate mfccs of each frame
    ppfFeaturesOfUnknownData = psf.mfcc(pfOnlySpeech, fSamplerate, numcep=iNumberOfMfccs, winlen=fVadFrameLengthInMS / 1000, winstep=fVadFrameLengthInMS / 1000 / 2)

    # Pick random frames of test signal with overall length of fLengthOfTestDataInSec
    ppfSplittedFeaturesOfUnknownData = splitTrainingData(ppfFeaturesOfUnknownData, fVadFrameLengthInMS, fLengthOfTestDataInSec)

    # Calculate most probable speaker model and store result
    piClassifiedData.append(classify_probability(ppfSplittedFeaturesOfUnknownData,list_models))

# Print estimated speaker for each file
for iIndex, iModel in enumerate(piClassifiedData):
    print('Estimated speaker for file ' + str(ppcFilenames[iIndex]) + ' is speaker: '+ str(iModel + 1))

# Plot first feature samples for each speaker model
if bPlotFeatureExample:
    plt.figure()
    plt.imshow(pppfTrainingData[0,:,:], aspect='auto', cmap=plt.get_cmap('jet'))

    plt.figure()
    plt.imshow(pppfTrainingData[1,:,:], aspect='auto', cmap=plt.get_cmap('jet'))

    plt.figure()
    plt.imshow(pppfTrainingData[2,:,:], aspect='auto', cmap=plt.get_cmap('jet'))

    plt.figure()
    plt.imshow(pppfTrainingData[3,:,:], aspect='auto', cmap=plt.get_cmap('jet'))

    plt.figure()
    plt.imshow(pppfTrainingData[4,:,:], aspect='auto', cmap=plt.get_cmap('jet'))

    plt.show()